import numpy as np
import matplotlib.pyplot as pp
import random as rd

## Initialisation
def lab(n):
    global tab
    # Pour que le tableau soit visible de l'extérieur et donc utilisable par les autres fonctions.
    tab=np.zeros((n,n),bool)
    # Tableau de taille (n,n) rempli de zéros booléens, ie de False.
    tab[0,0]=True
    # Elle est visitée car on y est au départ.
    
## Fonctions élémentaires      
def cocher(case):
    # Cette fonction permet de valider l'étape suivante quand on avance dans la pile. 
    x,y = case # extraction des coordonnées de case
    n=len(tab)
    if x in range(n) and y in range (n):
        tab[x,y]=True
        # on ne coche la case que si elle est bien dans le tableau.
 
def visitee(case):
    # Cette fonction permet de consulter si la case a été visitée, afin de ne pas revenir dessus ensuite.
    x,y = case
    n=len(tab)
    if x in range(n) and y in range (n):
        return tab[x,y]
    else:
        return True
        #  on marque les cases hors table comme visitées afin d'être sûr de ne pas sortir quand on atteint le bord.

def options(case):
    # Cette fonction donne la liste des cases disponibles parmi les 4 adjacentes à case.
    x,y = case
    adj=[(x-1,y),(x+1,y),(x,y-1),(x,y+1)]
    # liste des cases adjacentes
    op=[c for c in adj if not visitee(c)]
    # filtrage des cases pas encore visitées.
    return op
        
def choix(case):
    # Cette fonction choisit une case au hasard parmi les options de case.
    L=options(case)
    assert len(L)>0
    return rd.choice(L)
    # 'choice' renvoie un élément aléatoire d'une liste, avec équiprobabilité.
    
## Construction du labyrinthe    
# Le principe est de cheminer au hasard parmi les cases pas encore visitées.
# Si on arrive dans un cul de sac on revient en arrière jusqu'à ce qu'on ait des options disponibles.
# Tant qu'on peut avancer on relie les cases et on les met sur la pile. 
# Tant qu'il reste des options, c'est à dire des cases pas encore visitées, la pile n'est pas vide. Le test d'arrêt est donc la vacuité de la pile.

def labyr(n):
    lab(n)
    # On crée le labyrinthe logique.
    pp.close() 
    # Fermeture des précédents graphiques.
    pp.figure(figsize=(10,10), facecolor='b')
    # Prépare la fenêtre d'affichage: taille 10*10 pouces, fond noir.
    pp.axis('equal') 
    # Mêmes unités sur les 2 axes, pour que le labyrinthe soit carré. 
    pp.axis('off') 
    # Pas d'axes
    p=[(0,0)]
    # Initialisation de la pile de cheminement.
    while len(p)>0:
        case=p[-1]  # on regarde la dernière case explorée
        if options(case)==[]:
            p.pop()  # s'il n'y a pas d'option, on revient en arrière 
        else:  # s'il y a au moins 1 option :
            x0,y0 = case  # coordonnées de la position actuelle
            suiv=choix(case)  # case suivante
            x1,y1 = suiv  # ses coordonnées
            acces[x1,y1]=case  # on stocke dans le tableau acces la case qui a permis d'accéder à suiv  
            pp.plot([x0,x1],[y0,y1],'w',linewidth=200/n)  # on relie case et suiv par un trait bleu, de largeur inversement proportionnelle à n 
            cocher(suiv)  # car elle est visitée
            p.append(suiv)  # et on complète le chemin
    pp.show()
    # Affichage du tracé du labyrinthe.
    
## Mémoire des accès
# On va construire un matrice de taille (n,n) dont les éléments seront des tuples. En position (i,j) on stockera la case qui a précédé (i,j) au cours de l'empilement, dans la création du labyrinthe. 
# Comme cette pile évolue sans arrêt, la seule possibilité est de construire la matrice des accès en même temps que le labyrinthe. On modifie donc en ce sens la fonction labyr.

def labyr2(n):
    lab(n)
    global acces
    # La matrice des accès devra être visible de l'extérieur, pour pouvoir être exploitée par la fonction "solution".
    acces=np.empty((n,n),tuple) 
    # Matrice vide de taille (n,n) dont les éléments seront des tuples.
    pp.close()
    pp.figure(figsize=(12,11))
    pp.axis('equal')
    pp.axis('off')
    p=[(0,0)]
    pp.plot([-2,0],[0,0],'b',linewidth=300/n) # entrée du labyrinthe
    while len(p)>0:
        case=p[-1]  
        op=options(case)
        if options(case)==[]:
            p.pop()  
        else: 
            x0,y0 = case  
            suiv=choix(case)  
            x1,y1 = suiv  
            acces[x1,y1]=case  # on stocke dans le tableau acces la case qui a permis d'accéder à suiv  
            pp.plot([x0,x1],[y0,y1],'b',linewidth=300/n)
            cocher(suiv)  # car elle est visitée
            p.append(suiv)
    pp.plot([n-1,n+1],[n-1,n-1],'b',linewidth=300/n) # sortie du labyrinthe
    pp.show()
    
## Construction de la solution
# Pour trouver le chemin d'accés à la case fin, on récupère dans le tableau acces la case précédente, puis on recommence jusqu'à obtenir la case initiale.
# On trace ensuite la liste ainsi constituée.
# ATTENTION: pour que le tracé se fasse sur le labyrinthe déjà dessiné, il ne faut pas fermer la fenêtre d'affichage de ce labyrinthe.

def solution(fin):
    n=len(acces) # on récupère la taille du mabyrinthe
    case=fin # case initiale : on commence par la fin
    sol=[fin] # sera la pile du cheminement, à reculons
    x0,y0 = fin
    pp.plot([-2,0],[0,0],'g',linewidth=200/n) # entrée dans le labyrinthe
    while case!=(0,0): # on s'arrêtera quand on sera arrivé à la case de départ
        x,y = case
        prec=acces[x,y] # case précédente
        sol.append(prec) # on l'ajoute au cheminement
        case=prec # nouvelle case actuelle
    X=[c[0] for c in sol] # liste des abscisses des points de sol
    Y=[c[1] for c in sol] # liste des ordonnées des points de sol
    pp.plot(X,Y,'g',linewidth=200/n) # on trace en vert
    pp.plot([x0],[y0],'ro') # sortie du labyrinthe
    pp.show()